@@ -34,39 +34,38 @@
 ;;
 ;;;;; Code:
 
-;; we reuse several ada-mode functions
-(require 'ada-mode)
 (require 'cl-lib)
+(require 'gnat-core)
+(require 'gpr-indent-user-options)
+(require 'gpr-process)
+(require 'gpr-skel)
+(require 'wisi-process-parse)
+(require 'wisi-prj)
 
 (defgroup gpr nil
   "Major mode for editing gpr (Gnat Project File) source code in Emacs."
   :group 'languages)
 
-(defcustom gpr-process-parse-exec "gpr_mode_wisi_parse.exe"
-  "Name of executable to use for external process gpr parser,"
-  :type 'string
-  :group 'gpr)
-
 (defvar gpr-mode-map
   (let ((map (make-sparse-keymap)))
     ;; C-c <letter> are reserved for users
 
     ;; global-map has C-x ` 'next-error
-    (define-key map [return]   'ada-indent-newline-indent)
-    (define-key map "\C-c`"    'ada-show-secondary-error)
+    (define-key map [return]   'wisi-indent-newline-indent)
     ;; comment-dwim is in global map on M-;
-    (define-key map "\C-c\C-c" 'ada-build-make)
-    (define-key map "\C-c\C-e" 'gpr-expand)
-    (define-key map "\C-c\C-f" 'gpr-show-parse-error)
-    (define-key map "\C-c\C-i" 'gpr-indent-statement)
-    (define-key map "\C-c\C-o" 	 'ff-find-other-file)
+    (define-key map "\C-c\C-e" 'wisi-skel-expand)
+    (define-key map "\C-c\C-f" 'wisi-show-parse-error)
+    (define-key map "\C-c\C-i" 'wisi-indent-statement)
+    (define-key map "\C-c\C-o" 'ff-find-other-file)
     (define-key map "\C-c\C-P" 'gpr-set-as-project)
-    (define-key map "\C-c\C-t" 'ada-case-read-all-exceptions)
-    (define-key map "\C-c\C-w" 'ada-case-adjust-at-point)
-    (define-key map "\C-c\C-y" 'ada-case-create-exception)
-    (define-key map "\C-c\C-\M-y" 'ada-case-create-partial-exception)
-    (define-key map "\M-n" 'skeleton-next-placeholder)
-    (define-key map "\M-p" 'skeleton-prev-placeholder)
+    (define-key map "\C-c\C-w" 'wisi-case-adjust-at-point)
+    (define-key map "\C-c\C-y" 'wisi-case-create-exception)
+    (define-key map "\C-c\C-\M-y" 'wisi-case-create-partial-exception)
+    (define-key map "\M-n" 'wisi-skel-next-placeholder)
+    (define-key map "\M-p" 'wisi-skel-prev-placeholder)
+
+    (wisi-case-activate-keys map)
+
     map
   )  "Local keymap used for GPR mode.")
 
@@ -80,64 +79,25 @@
      ["Key bindings"          describe-bindings t]
      )
 
-    ["Customize"     (customize-group 'ada)];; we reuse the Ada indentation options
+    ["Customize"     (customize-group 'gpr)]
     ["------"        nil nil]
-    ["Build current project"       ada-build-make                   t]
-    ["Find and select project ..." ada-build-prompt-select-prj-file t]
-    ["Select project ..."          ada-prj-select                   t]
-    ["Parse and select current file" gpr-set-as-project             t]
-    ["Show current project"        ada-prj-show                     t]
-    ["Show project search path"    ada-prj-show-prj-path            t]
+    ["Show current project"        wisi-prj-show                    t]
+    ["Show project search path"    wisi-prj-show-prj-path           t]
     ["Next compilation error"      next-error                       t]
-    ["Show secondary error"        ada-show-secondary-error         t]
-    ["Show last parse error"       gpr-show-parse-error             t]
+    ["Show last parse error"       wisi-show-parse-error            t]
     ["Other file"                  ff-find-other-file               t]
     ("Edit"
      ["Indent Line or selection"      indent-for-tab-command         t]
-     ["Indent current statement"      gpr-indent-statement           t]
+     ["Indent current statement"      wisi-indent-statement          t]
      ["Indent Lines in File"          (indent-region (point-min) (point-max))  t]
-     ["Expand skeleton"               gpr-expand                     t]
-     ["Next skeleton placeholder"     skeleton-next-placeholder      t]
-     ["Previous skeleton placeholder" skeleton-prev-placeholder      t]
+     ["Expand skeleton"               wisi-skel-expand               t]
+     ["Next skeleton placeholder"     wisi-skel-next-placeholder     t]
+     ["Previous skeleton placeholder" wisi-skel-prev-placeholder     t]
      ["Comment/uncomment selection"   comment-dwim                   t]
      ["Fill Comment Paragraph"        fill-paragraph                 t]
-
-     ["Fill Comment Paragraph Justify" ada-fill-comment-paragraph-justify t]
-     ["Fill Comment Paragraph Postfix" ada-fill-comment-paragraph-postfix t]
      )
     ))
 
-(defvar gpr-show-parse-error nil
-  ;; Supplied by indentation engine parser
-  "Function to show last error reported by indentation parser."
-  )
-
-(defun gpr-show-parse-error ()
-  (interactive)
-  (when gpr-show-parse-error
-    (funcall gpr-show-parse-error)))
-
-(defvar gpr-expand nil
-  ;; skeleton function
-  "Function to call to expand tokens (ie insert skeletons).")
-
-(defun gpr-expand ()
-  "Expand previous word into a statement skeleton."
-  (interactive)
-  (when gpr-expand
-    (funcall gpr-expand)))
-
-(defvar gpr-indent-statement nil
-  ;; indentation function
-  "Function to indent the statement/declaration point is in or after.
-Function is called with no arguments.")
-
-(defun gpr-indent-statement ()
-  "Indent current statement."
-  (interactive)
-  (when gpr-indent-statement
-    (funcall gpr-indent-statement)))
-
 (defconst gpr-keywords
   '(
     "abstract"
@@ -156,7 +116,7 @@ Function is called with no arguments.")
     "null"
     "others"
     "package"
-    "project"
+    ;; "project" may also be a non-keyword attribute prefix; see test/gpr/gds.gpr
     "renames"
     "standard"
     "type"
@@ -174,7 +134,6 @@ Function is called with no arguments.")
   "Expressions to highlight in gpr mode.")
 
 (defun gpr-ff-special-with ()
-  (ada-require-project-file)
   (let ((project-path (match-string 1)))
     ;; project-path may be any of "foo", "foo.gpr", "../foo.gpr"
     ;;
@@ -200,15 +159,57 @@ Function is called with no arguments.")
 	       'gpr-ff-special-with)
 	 )))
 
-(defvar gpr-which-function nil
-  ;; supplied by the indentation engine
-  "Function called with no parameters; it should return the name
-of the package or project point is in or just after, or nil.")
-
 (defun gpr-which-function ()
-  "See `gpr-which-function' variable."
-  (when gpr-which-function
-    (funcall gpr-which-function)))
+  "Return the name of the package or project point is in or just after, or nil."
+  (wisi-validate-cache (point-min) (point) nil 'navigate)
+  ;; No message on parse fail, since this could be called from which-function-mode
+  (when (wisi-cache-covers-pos 'navigate (point))
+    (let ((cache (wisi-backward-cache))
+	  done
+	  project-pos
+	  package-pos
+	  decl-pos)
+      (while (and cache (not done))
+	;; find attribute_declaration and package containing point (if any)
+	(cond
+	 ((not (eq (wisi-cache-class cache) 'statement-start))
+	  nil)
+
+	 ((eq (wisi-cache-nonterm cache) 'attribute_declaration)
+	  (setq decl-pos (point)))
+
+	 ((eq (wisi-cache-nonterm cache) 'package_spec)
+	  (setq package-pos (point))
+	  (setq done t))
+
+	 ((eq (wisi-cache-nonterm cache) 'simple_project_declaration)
+	  (setq project-pos (point))
+	  (setq done t))
+	 )
+
+	(setq cache (wisi-goto-containing cache)))
+
+      (cond
+       (package-pos
+	(goto-char package-pos)
+	(setq done t))
+
+       (decl-pos
+	(goto-char decl-pos)
+	(setq done t))
+
+       (project-pos
+	(goto-char project-pos)
+	(setq done t))
+
+       (t ;; before project
+	(setq done nil))
+       )
+
+      (when done
+	(wisi-next-name))
+
+      )))
 
 (defun gpr-add-log-current-function ()
   "For `add-log-current-defun-function'. Returns enclosing package or project name."
@@ -220,19 +221,6 @@ of the package or project point is in or just after, or nil.")
     (end-of-line 1)
     (gpr-which-function)))
 
-(declare-function gpr-query-kill-all-sessions "gpr-query.el" nil)
-(defun gpr-set-as-project (&optional file)
-  "Set FILE (default current buffer file) as Emacs project file."
-  (interactive)
-  (save-some-buffers t)
-  ;; Kill sessions to catch changed env vars
-  (cl-ecase ada-xref-tool
-    (gnat nil)
-    (gpr_query (gpr-query-kill-all-sessions))
-    )
-  (ada-parse-prj-file (or file (buffer-file-name)))
-  (ada-select-prj-file (or file (buffer-file-name))))
-
 (defvar gpr-mode-syntax-table
   (let ((table (make-syntax-table)))
     ;; (info "(elisp)Syntax Class Table" "*info syntax class table*")
@@ -247,13 +235,14 @@ of the package or project point is in or just after, or nil.")
     (modify-syntax-entry ?\" "\"" table)
 
     ;; punctuation; operators etc
+    (modify-syntax-entry ?-  ". 124" table); operator, double hyphen as comment
     (modify-syntax-entry ?&  "." table)
     (modify-syntax-entry ?. "." table)
     (modify-syntax-entry ?:  "." table)
     (modify-syntax-entry ?=  "." table)
     (modify-syntax-entry ?>  "." table)
     (modify-syntax-entry ?\; "." table)
-    (modify-syntax-entry ?\\ "." table); default is escape; not correct for Ada strings
+    (modify-syntax-entry ?\\ "." table); default is escape; not correct for gpr strings
     (modify-syntax-entry ?\|  "." table)
 
     ;; and \f and \n end a comment
@@ -265,7 +254,7 @@ of the package or project point is in or just after, or nil.")
     (modify-syntax-entry ?\( "()" table)
     (modify-syntax-entry ?\) ")(" table)
 
-    ;; skeleton placeholder delimiters; see ada-skel.el. We use generic
+    ;; skeleton placeholder delimiters; see gpr-skel.el. We use generic
     ;; comment delimiter class, not comment starter/comment ender, so
     ;; these can be distinguished from line end.
     (modify-syntax-entry ?{ "!" table)
@@ -275,21 +264,93 @@ of the package or project point is in or just after, or nil.")
     )
   "Syntax table to be used for editing gpr source code.")
 
-(defun gpr-syntax-propertize (start end)
-  "Assign `syntax-table' properties in accessible part of buffer."
-  ;; (info "(elisp)Syntax Properties")
-  ;;
-  ;; called from `syntax-propertize', inside save-excursion with-silent-modifications
-  ;; syntax-propertize-extend-region-functions is set to
-  ;; syntax-propertize-wholelines by default.
-  (let ((inhibit-read-only t)
-	(inhibit-point-motion-hooks t))
-    (goto-char start)
-    (save-match-data
-      (while (re-search-forward "--" end t); comment start
-	(put-text-property
-	 (match-beginning 0) (match-end 0) 'syntax-table '(11 . nil)))
-  )))
+;;;; wisi integration
+
+(defcustom gpr-auto-case t
+  "When non-nil, automatically change case of preceding word while
+typing.  Casing of gpr keywords is done according to `gpr-case-keyword',
+identifiers according to `gpr-case-identifier'."
+  :group 'gpr
+  :type  '(choice (const nil)
+		  (const t))
+  :safe  (lambda (val) (memq val '(nil t))))
+(make-variable-buffer-local 'gpr-auto-case)
+
+(defcustom gpr-case-keyword 'lower-case
+  "Indicate how to adjust case for language keywords.
+Value is one of lower-case, upper-case."
+  :group 'gpr
+  :type '(choice (const lower-case)
+		 (const upper-case))
+  :safe #'symbolp)
+(make-variable-buffer-local 'gpr-case-keyword)
+
+(defcustom gpr-case-strict t
+  "If non-nil, force Mixed_Case for identifiers.
+Otherwise, allow UPPERCASE for identifiers."
+  :group 'gpr
+  :type 'boolean
+  :safe  #'booleanp)
+(make-variable-buffer-local 'gpr-case-strict)
+
+(defcustom gpr-case-identifier 'mixed-case
+  "Indicates how to adjust the case of gpr keywords."
+  :group 'gpr
+  :type '(choice (const mixed-case)
+		 (const lower-case)
+		 (const upper-case))
+  ;; see comment on :safe at gpr-case-keyword
+  :safe (lambda (val) (memq val '(mixed-case lower-case upper-case))))
+(make-variable-buffer-local 'gpr-case-identifier)
+
+(defun gpr-case-adjust-p (_typed-char)
+  "For `wisi-case-adjust-p-function'."
+  ;; casing of 'project' in "Project'Project_Dir" vs "project GDS is"
+  (save-excursion
+    (let ((end (1+ (point)))
+	  (start (progn (skip-syntax-backward "w")(point))))
+      (cond
+       ((string= "project" (downcase (buffer-substring-no-properties start end)))
+	(cond
+	 ((= (char-after end) ?') ;; attribute Project'
+	  t)
+	 ((= (char-after end) ? ) ;; keyword project GDS
+	  (wisi-case-keyword start end)
+	  nil)
+	 ))
+
+       (t t) ;; not "project"
+       ))))
+
+(cl-defstruct (gpr-wisi-parser (:include wisi-process--parser))
+  ;; no new slots
+  )
+
+(cl-defstruct (gpr-prj (:include wisi-prj))
+  ;; no new slots
+  )
+
+(defun gpr-prj-default (&optional name)
+  (make-gpr-prj :name (or name "_default_") :compiler (make-gnat-compiler)))
+
+(cl-defmethod wisi-prj-default ((prj gpr-prj))
+  (gpr-prj-default (wisi-prj-name prj)))
+
+(cl-defmethod wisi-parse-format-language-options ((_parser gpr-wisi-parser))
+  (format "%d %d %d"
+	  gpr-indent
+	  gpr-indent-broken
+	  gpr-indent-when
+	  ))
+
+(defconst gpr-wisi-language-protocol-version "1"
+  "Defines language-specific parser parameters.
+Must match wisi-gpr.ads Language_Protocol_Version.")
+
+(defcustom gpr-process-parse-exec "gpr_mode_wisi_parse.exe"
+  "Name of executable to use for external process gpr parser,"
+  :type 'string
+  :group 'gpr)
 
 ;;;;
 ;;;###autoload
@@ -302,10 +363,6 @@ of the package or project point is in or just after, or nil.")
   (setq mode-name "GNAT Project")
   (use-local-map gpr-mode-map)
   (set-syntax-table gpr-mode-syntax-table)
-  (set (make-local-variable 'syntax-propertize-function) 'gpr-syntax-propertize)
-  (when (boundp 'syntax-begin-function)
-    ;; obsolete in emacs-25.1
-    (set (make-local-variable 'syntax-begin-function) nil))
   (set 'case-fold-search t); gpr is case insensitive; the syntax parsing requires this setting
   (set (make-local-variable 'comment-start) "--")
   (set (make-local-variable 'comment-end) "")
@@ -314,9 +371,6 @@ of the package or project point is in or just after, or nil.")
 
   (set (make-local-variable 'require-final-newline) t)
 
-  (ada-case-activate-keys gpr-mode-map)
-  (set (make-local-variable 'ada-keywords) gpr-keywords)
-
   (set (make-local-variable 'font-lock-defaults)
        '(gpr-font-lock-keywords
 	 nil t
@@ -328,8 +382,28 @@ of the package or project point is in or just after, or nil.")
   (set (make-local-variable 'add-log-current-defun-function)
        'gpr-add-log-current-function)
 
+  (wisi-setup
+   :indent-calculate nil
+   :post-indent-fail nil
+   :parser
+   (wisi-process-parse-get
+    (make-gpr-wisi-parser
+     :label "gpr"
+     :language-protocol-version gpr-wisi-language-protocol-version
+     :exec-file gpr-process-parse-exec
+     :face-table gpr-process-face-table
+     :token-table gpr-process-token-table
+     :repair-image gpr-process-repair-image
+     )))
+
   (run-mode-hooks 'gpr-mode-hook)
 
+  (setq wisi-auto-case gpr-auto-case)
+  (setq wisi-case-identifier gpr-case-identifier)
+  (setq wisi-case-strict gpr-case-strict)
+  (setq wisi-language-keywords gpr-keywords)
+  (setq wisi-case-keyword gpr-case-keyword)
+  (setq wisi-case-adjust-p-function #'gpr-case-adjust-p)
   )
 
 ;;;###autoload
@@ -337,29 +411,12 @@ of the package or project point is in or just after, or nil.")
 
 (put 'gpr-mode 'custom-mode-group 'gpr)
 
-(defvar gpr-parser nil
+(defvar gpr-parser 'process
   "Indicate parser and lexer to use for gpr buffers:
 
-elisp : wisi parser and lexer implemented in elisp.
-
 process : wisi elisp lexer, external process parser specified
   by ‘gpr-process-parse-exec ’.
 ")
 
 (provide 'gpr-mode)
-
-(require 'gpr-wisi)
-
-(cl-case gpr-parser
-  (elisp nil)
-  (process nil)
-  (t
-   (if (locate-file gpr-process-parse-exec exec-path '("" ".exe"))
-       (setq gpr-parser 'process)
-     (setq gpr-parser 'elisp)))
-  )
-
-(unless (featurep 'gpr-skeletons)
-  (require 'gpr-skel))
-
-;;; end of file
+;;; gpr-mode.el ends here
