@@ -28,16 +28,27 @@
 ;;
 ;; M-x gpr-query
 
-(require 'ada-mode-compat) ;; font-lock-ensure
-
-(require 'ada-mode) ;; for ada-prj-*, some other things
-(require 'gnat-core)
 (require 'cl-lib)
 (require 'compile)
+(require 'gnat-core)
+(require 'xref)
+(require 'wisi-prj)
 
-(eval-and-compile
-  (when (> emacs-major-version 24)
-    (require 'xref)))
+(defgroup gpr-query nil
+  "gpr_query cross reference tool"
+  :group 'tools)
+
+(defcustom gpr-query-exec "gpr_query"
+  "Executable for gpr_query."
+  :type 'string)
+
+(defcustom gpr-query-env nil
+  "Environment variables needed by the gpr_query executable.
+Value must be alist where each element is \"<name>=<value>\""
+  ;; This could also be provided as a project file setting, But it is
+  ;; intended for LD_LIBRARY_PATH (info "(ada-mode)Ada executables"),
+  ;; which must be set for all projects on the system.
+  :type 'string)
 
 ;;;;; sessions
 
@@ -61,28 +72,42 @@
 ;; *xref* buffer.
 (defconst gpr-query-buffer-name-prefix "*gpr_query-")
 
-(defun gpr-query--start-process (session)
+(defvar gpr-query--debug-start nil)
+
+(defun gpr-query--start-process (project session)
   "Start the session process running gpr_query."
   (unless (buffer-live-p (gpr-query--session-buffer session))
     ;; user may have killed buffer
-    (setf (gpr-query--session-buffer session) (gnat-run-buffer gpr-query-buffer-name-prefix))
+    (setf (gpr-query--session-buffer session)
+	  (gnat-run-buffer project (gnat-compiler-run-buffer-name (wisi-prj-xref project))))
     (with-current-buffer (gpr-query--session-buffer session)
       (compilation-mode)
       (setq buffer-read-only nil)))
 
   (with-current-buffer (gpr-query--session-buffer session)
-    (let ((process-environment (cl-copy-list (ada-prj-get 'proc_env)))
-	  ;; for GPR_PROJECT_PATH, other env vars set in ada-mode
-	  ;; project files and used by gpr files.
-
-	  (project-file (file-name-nondirectory (ada-prj-get 'gpr_file))))
+    (let ((process-environment
+	   (append
+	    (wisi-prj-compile-env project)
+	    (wisi-prj-file-env project)
+	    gpr-query-env
+	    (copy-sequence process-environment)))
+	  (gpr-file (file-name-nondirectory (gnat-compiler-gpr-file (wisi-prj-xref project)))))
 
       (erase-buffer); delete any previous messages, prompt
       (setf (gpr-query--session-process session)
-	    (start-process (concat "gpr_query " (buffer-name))
-			   (gpr-query--session-buffer session)
-			   "gpr_query"
-			   (concat "--project=" project-file)))
+	    (apply #'start-process
+		   (concat "gpr_query " (buffer-name))
+		   (gpr-query--session-buffer session)
+		   gpr-query-exec
+		   (cl-delete-if
+		    'null
+		    (list
+		     (concat "--project=" gpr-file)
+		     (when (not ada-xref-full-path) "--short_file_names")
+		     (when gpr-query--debug-start
+		       (concat "--tracefile=gpr_query.trace")
+		       ;; The file gpr_query.trace should contain: gpr_query=yes
+		       )))))
       (set-process-query-on-exit-flag (gpr-query--session-process session) nil)
       (gpr-query-session-wait session)
 
@@ -101,31 +126,31 @@
 	(error "gpr-query process failed to start"))
       )))
 
-(defun gpr-query--make-session ()
+(defun gpr-query--make-session (project)
   "Create and return a session for the current project file."
   (let ((session
 	 (make-gpr-query--session
 	  :buffer nil
 	  :process nil)))
-    (gpr-query--start-process session)
+    (gpr-query--start-process project session)
     session))
 
 (defvar gpr-query--sessions '()
   "Assoc list of sessions, indexed by absolute GNAT project file name.")
 
-(defun gpr-query-cached-session ()
-  "Return a session for the current project file, creating it if necessary."
-  (let ((session (cdr (assoc ada-prj-current-file gpr-query--sessions))))
+(defun gpr-query-cached-session (project)
+  "Return a session for PROJECT, creating it if necessary."
+  (let* ((gpr-file (gnat-compiler-gpr-file (wisi-prj-xref project)))
+	 (session (cdr (assoc gpr-file gpr-query--sessions))))
     (if session
 	(progn
 	  (unless (process-live-p (gpr-query--session-process session))
-	    (gpr-query--start-process session))
+	    (gpr-query--start-process project session))
 	  session)
       ;; else
       (prog1
-          (setq session (gpr-query--make-session))
-	(setq gpr-query--sessions
-	      (cl-acons ada-prj-current-file session gpr-query--sessions))))
+          (setq session (gpr-query--make-session project))
+	(push (cons gpr-file session) gpr-query--sessions)))
     ))
 
 (defconst gpr-query-prompt "^>>> $"
@@ -158,29 +183,21 @@
 	(error "gpr_query process died"))
       )))
 
-(defun gpr-require-prj ()
-  "Throw error if no project file defined."
-  (unless (or (ada-prj-get 'gpr_file)
-	      (ada-prj-get 'gpr_query_file))
-    (error "no gpr project file defined.")))
-
-(defun gpr-query-session-send (cmd wait)
-  "Send CMD to gpr_query session for current project.
+(defun gpr-query-session-send (session cmd wait)
+  "Send CMD to gpr_query session for GPR-FILE.
 If WAIT is non-nil, wait for command to complete.
 Return buffer that holds output."
-  (gpr-require-prj)
-  (let ((session (gpr-query-cached-session)))
-    ;; always wait for previous command to complete; also checks for
-    ;; dead process.
-    (gpr-query-session-wait session)
-    (with-current-buffer (gpr-query--session-buffer session)
-      (erase-buffer)
-      (process-send-string (gpr-query--session-process session)
-			   (concat cmd "\n"))
-      (when wait
-	(gpr-query-session-wait session))
-      (current-buffer)
-      )))
+  ;; always wait for previous command to complete; also checks for
+  ;; dead process.
+  (gpr-query-session-wait session)
+  (with-current-buffer (gpr-query--session-buffer session)
+    (erase-buffer)
+    (process-send-string (gpr-query--session-process session)
+			 (concat cmd "\n"))
+    (when wait
+      (gpr-query-session-wait session))
+    (current-buffer)
+    ))
 
 (defun gpr-query-kill-session (session)
   (let ((process (gpr-query--session-process session)))
@@ -203,55 +220,55 @@ Return buffer that holds output."
     (message "Killed %d sessions" count)
     ))
 
-(defun gpr-query-show-buffer (&optional session)
-  "For `ada-show-xref-tool-buffer'; show gpr-query buffer for current project."
+(defun gpr-query-show-buffer (session)
   (interactive)
-  (pop-to-buffer (gpr-query--session-buffer (or session (gpr-query-cached-session)))))
+  (pop-to-buffer (gpr-query--session-buffer session)))
 
 ;;;;; utils
 
-(defun gpr-query-get-src-dirs (src-dirs)
-  "Append list of source dirs in current gpr project to SRC-DIRS.
+(defun gpr-query-get-src-dirs (project src-dirs)
+  "Append list of source dirs in gpr project PROJECT to SRC-DIRS.
 Uses `gpr_query'. Returns new list."
 
-  (with-current-buffer (gpr-query--session-buffer (gpr-query-cached-session))
-    (gpr-query-session-send "source_dirs" t)
-    (goto-char (point-min))
-    (while (not (looking-at gpr-query-prompt))
-      (cl-pushnew
-       (expand-file-name ; Canonicalize path part.
-	(directory-file-name
-	 (buffer-substring-no-properties (point) (point-at-eol))))
-	src-dirs :test #'equal)
-      (forward-line 1))
-    )
+  (let ((session (gpr-query-cached-session project)))
+    (with-current-buffer (gpr-query-session-send session "source_dirs" t)
+      (goto-char (point-min))
+      (while (not (looking-at gpr-query-prompt))
+	(cl-pushnew
+	 (expand-file-name ; Canonicalize path part.
+	  (directory-file-name
+	   (buffer-substring-no-properties (point) (point-at-eol))))
+	 src-dirs :test #'equal)
+	(forward-line 1))
+      ))
   src-dirs)
 
-(defun gpr-query-get-prj-dirs (prj-dirs)
-  "Append list of project dirs in current gpr project to PRJ-DIRS.
+(defun gpr-query-get-prj-dirs (project prj-dirs)
+  "Append list of project dirs in gpr project PROJECT to PRJ-DIRS.
 Uses `gpr_query'. Returns new list."
 
-  (with-current-buffer (gpr-query--session-buffer (gpr-query-cached-session))
-    (gpr-query-session-send "project_path" t)
-    (goto-char (point-min))
-    (while (not (looking-at gpr-query-prompt))
-      (cl-pushnew
-       (let ((dir (buffer-substring-no-properties (point) (point-at-eol))))
-	 (if (string= dir ".")
-	     (directory-file-name default-directory)
-	   (expand-file-name dir))) ; Canonicalize path part.
-       prj-dirs
-       :test #'equal)
-      (forward-line 1))
-    )
+  (let ((session (gpr-query-cached-session project)))
+    (with-current-buffer (gpr-query-session-send session "project_path" t)
+      (goto-char (point-min))
+      (while (not (looking-at gpr-query-prompt))
+	(cl-pushnew
+	 (let ((dir (buffer-substring-no-properties (point) (point-at-eol))))
+	   (if (string= dir ".")
+	       (directory-file-name default-directory)
+	     (expand-file-name dir))) ; Canonicalize path part.
+	 prj-dirs
+	 :test #'equal)
+	(forward-line 1))
+      ))
   prj-dirs)
 
 (defconst gpr-query-ident-file-regexp
   ;; C:\Projects\GDS\work_dscovr_release\common\1553\gds-mil_std_1553-utf.ads:252:25
   ;; /Projects/GDS/work_dscovr_release/common/1553/gds-mil_std_1553-utf.ads:252:25
-  "\\(\\(?:.:\\\\\\|/\\)[^:]*\\):\\([0123456789]+\\):\\([0123456789]+\\)"
+  ;; gds-mil_std_1553-utf.ads:252:25 - when ada-xref-full-path is nil
+  "\\(\\(?:.:\\\\\\|/\\)?[^:]*\\):\\([0123456789]+\\):\\([0123456789]+\\)"
   ;; 1                             2                   3
-  "Regexp matching <file>:<line>:<column>")
+  "Regexp matching <file>:<line>:<column> where <file> is an absolute file name or basename.")
 
 (defconst gpr-query-ident-file-regexp-alist
   (list (concat "^" gpr-query-ident-file-regexp) 1 2 3)
@@ -261,31 +278,49 @@ Uses `gpr_query'. Returns new list."
   (concat gpr-query-ident-file-regexp " (\\(.*\\))")
   "Regexp matching <file>:<line>:<column> (<type>)")
 
-(defun gpr-query-compilation (identifier file line col cmd comp-err)
+(defun gpr-query-compilation (project identifier file line col cmd comp-err &optional local_only append)
   "Run gpr_query IDENTIFIER:FILE:LINE:COL CMD,
 with compilation-error-regexp-alist set to COMP-ERR."
   ;; Useful when gpr_query will return a list of references; the user
   ;; can navigate to each result in turn via `next-error'.
-  ;; FIXME: implement ada-xref-full-path.
-  ;;
-  ;; FIXME: implement append
+
+  ;; Note that ada-xref-full-path is used in gpr-query--start-process,
+  ;; not here; if it changes, the gpr-query session must be restarted.
 
   ;; Emacs column is 0-indexed, gpr_query is 1-indexed.
-  (let ((cmd-1 (format "%s %s:%s:%d:%d" cmd identifier file line (1+ col)))
-	(result-count 0)
-	target-file target-line target-col)
-    (with-current-buffer (gpr-query--session-buffer (gpr-query-cached-session))
-      (setq buffer-read-only nil)
-      (set (make-local-variable 'compilation-error-regexp-alist) (list comp-err))
-      (gpr-query-session-send cmd-1 t)
+  (let* ((cmd-1 (format "%s %s:%s:%d:%d" cmd identifier file line (1+ col)))
+	 (cmd-2 (if local_only (concat cmd-1 " local_only") cmd-1))
+	 (session (gpr-query-cached-session project))
+	 (result-count 0)
+	 start-pos prev-content
+	 target-file target-line target-col)
 
+    (when append
+      (with-current-buffer (gpr-query--session-buffer session)
+	;; don't include trailing prompt in `prev-content'
+	(goto-char (point-max))
+	(forward-line 0)
+	(setq prev-content (buffer-substring (point-min) (point)))))
+
+    (with-current-buffer (gpr-query-session-send session cmd-2 t)
       ;; point is at EOB. gpr_query returns one line per result plus prompt, warnings
       (setq result-count (- (line-number-at-pos) 1))
+      (setq start-pos (point-min))
+
+      (setq buffer-read-only nil)
+      (when append
+	(goto-char (point-min))
+	(insert prev-content)
+	(setq start-pos (point))
+	(goto-char (point-max)))
+
+      (set (make-local-variable 'compilation-error-regexp-alist) (list comp-err))
 
       (compilation--flush-parse (point-min) (point-max))
       (compilation--ensure-parse (point-max))
 
-      (goto-char (point-min))
+      (goto-char start-pos)
+
       (cond
        ((looking-at "^warning: ")
 	(setq result-count (1- result-count))
@@ -294,33 +329,36 @@ with compilation-error-regexp-alist set to COMP-ERR."
 	(error (buffer-substring-no-properties (line-beginning-position) (line-end-position))))
        )
 
-      (cl-case result-count
-	(0
-	 (error "gpr_query returned no results"))
-	(1
-	 ;; just go there, don't display session-buffer. We have to
-	 ;; fetch the compilation-message while in the
-	 ;; session-buffer. and call ada-goot-source outside the
-	 ;; with-current-buffer above.
-	 (let* ((msg (compilation-next-error 0))
-                ;; IMPROVEME: '--' indicates internal-only. But we can't
-                ;; use compile-goto-error, because that displays the
-                ;; session-buffer.
-	 	(loc (compilation--message->loc msg)))
-	   (setq target-file (caar (compilation--loc->file-struct loc))
-		 target-line (caar (cddr (compilation--loc->file-struct loc)))
-		 target-col  (1- (compilation--loc->col loc))
-		 )
-	   ))
+      (cond
+       ((and (not append)
+	     (= result-count 1))
+	;; just go there, don't display session-buffer. We have to
+	;; fetch the compilation-message while in the
+	;; session-buffer. and call wisi-goto-source outside the
+	;; with-current-buffer above.
+	(let* ((msg (compilation-next-error 0))
+               ;; IMPROVEME: '--' indicates internal-only. But we can't
+               ;; use compile-goto-error, because that displays the
+               ;; session-buffer.
+	       (loc (compilation--message->loc msg)))
+	  (setq target-file (caar (compilation--loc->file-struct loc))
+		target-line (caar (cddr (compilation--loc->file-struct loc)))
+		target-col  (1- (compilation--loc->col loc))
+		)
+	  ))
 
-	(t
-	 ;; for next-error, below
-	 (setq next-error-last-buffer (current-buffer)))
+       ((= result-count 0)
+	(error "gpr_query returned no results"))
 
-	));; case, with-currrent-buffer
+       (t
+	;; for next-error, below
+	(setq next-error-last-buffer (current-buffer)))
 
-    (if (= result-count 1)
-	(ada-goto-source target-file target-line target-col)
+       ));; case, with-currrent-buffer
+
+    (if (and (not append)
+	     (= result-count 1))
+	(wisi-goto-source target-file target-line target-col)
 
       ;; more than one result; display session buffer, goto first ref
       ;;
@@ -328,10 +366,11 @@ with compilation-error-regexp-alist set to COMP-ERR."
       ;; at point-min; work around that by moving forward 0 errors for
       ;; the first one. Unless the first line contains "warning: ".
       (pop-to-buffer next-error-last-buffer)
-      (goto-char (point-min))
-      (if (looking-at "^warning: ")
-	  (next-error 1 t)
-	(next-error 0 t))
+      (goto-char start-pos)
+      (unless append
+	(if (looking-at "^warning: ")
+	    (next-error 1 t)
+	  (next-error 0 t)))
       )
     ))
 
@@ -340,69 +379,11 @@ with compilation-error-regexp-alist set to COMP-ERR."
   (+ (abs (- found-col col))
      (* (abs (- found-line line)) 250)))
 
-;;;;; user interface functions
-
-(defun gpr-query-show-references ()
-  "Show all references of identifier at point."
-  (interactive)
-  (gpr-query-all
-   (thing-at-point 'symbol)
-   (file-name-nondirectory (buffer-file-name))
-   (line-number-at-pos)
-   (current-column))
-  )
-
-(defun gpr-query-overridden ()
-  "Move to the overridden declaration of the identifier around point.
-If OTHER-WINDOW (set by interactive prefix) is non-nil, show the
-buffer in another window."
-  (interactive "P")
-
-  (let ((target
-	 (gpr-query-overridden-1
-	  (thing-at-point 'symbol)
-	  (buffer-file-name)
-	  (line-number-at-pos)
-	  (save-excursion
-	    (goto-char (car (bounds-of-thing-at-point 'symbol)))
-	    (current-column))
-	  )))
-
-    (ada-goto-source (nth 0 target)
-		     (nth 1 target)
-		     (nth 2 target))
-    ))
-
-(defun gpr-query-goto-declaration ()
-  "Move to the declaration or body of the identifier around point.
-If at the declaration, go to the body, and vice versa. If at a
-reference, goto the declaration.
-
-If OTHER-WINDOW (set by interactive prefix) is non-nil, show the
-buffer in another window."
-  (interactive "P")
-
-  (let ((target
-	 (gpr-query-other
-	  (thing-at-point 'symbol)
-	  (buffer-file-name)
-	  (line-number-at-pos)
-	  (save-excursion
-	    (goto-char (car (bounds-of-thing-at-point 'symbol)))
-	    (current-column))
-	  )))
-
-    (ada-goto-source (nth 0 target)
-		     (nth 1 target)
-		     (nth 2 target))
-    ))
-
 (defvar gpr-query-map
   (let ((map (make-sparse-keymap)))
     ;; C-c C-i prefix for gpr-query minor mode
 
     (define-key map "\C-c\C-i\C-d" 'gpr-query-goto-declaration)
-    (define-key map "\C-c\C-i\C-p" 'ada-build-prompt-select-prj-file)
     (define-key map "\C-c\C-i\C-q" 'gpr-query-refresh)
     (define-key map "\C-c\C-i\C-r" 'gpr-query-show-references)
     ;; IMPROVEME: (define-key map "\C-c\M-d" 'gpr-query-parents)
@@ -413,18 +394,14 @@ buffer in another window."
 (defvar gpr-query-menu (make-sparse-keymap "gpr-query"))
 (easy-menu-define gpr-query-menu gpr-query-map "Menu keymap for gpr-query minor mode"
   '("gpr-query"
-    ["Find and select project ..."   ada-build-prompt-select-prj-file t]
-    ["Select project ..."            ada-prj-select                   t]
-    ["Show current project"          ada-prj-show                     t]
-    ["Show gpr-query buffer"         gpr-query-show-buffer            t]
-    ["Next compilation error"        next-error                       t]
-    ["Show secondary error"          ada-show-secondary-error         t]
-    ["Goto declaration/body"         gpr-query-goto-declaration       t]
-    ["Show parent declarations"      ada-show-declaration-parents     t]
-    ["Show references"               gpr-query-show-references        t]
-    ;; ["Show overriding"               gpr-query-show-overriding        t]
-    ;; ["Show overridden"               gpr-query-show-overridden        t]
-    ["Refresh cross reference cache" gpr-query-refresh        t]
+    ["Show gpr-query buffer"         gpr-query-show-buffer 	   t]
+    ["Next xref"                     next-error 		   t]
+    ["Goto declaration/body"         xref-find-definitions 	   t]
+    ["Show parent declarations"      wisi-show-declaration-parents t]
+    ["Show references"               wisi-show-references 	   t]
+    ["Show overriding"               wisi-show-overriding 	   t]
+    ["Show overridden"               wisi-show-overridden 	   t]
+    ["Refresh cross reference cache" wisi-refresh-prj-cache 	   t]
     ))
 
 (define-minor-mode gpr-query
@@ -436,52 +413,115 @@ Enable mode if ARG is positive."
   ;; just enable the menu and keymap
   )
 
-;;;;; support for Ada mode
+(defun gpr-query--normalize-filename (file &optional dir)
+  "Takes account of filesystem differences."
+  ;; need DIR to handle ada-xref-full-path nil
+  (cond
+   ((eq system-type 'windows-nt)
+    ;; 'expand-file-name' converts Windows directory
+    ;; separators to normal Emacs.  Since Windows file
+    ;; system is case insensitive, GNAT and Emacs can
+    ;; disagree on the case, so convert all to lowercase.
+    (downcase (expand-file-name file dir)))
 
-(defun gpr-query-refresh (delete-files)
-  "For `ada-xref-refresh-function', using gpr_query."
-  (interactive "P")
+   ((eq system-type 'darwin)
+    ;; case-insensitive case-preserving; so just downcase
+    (when (and dir
+	       (not ada-xref-full-path))
+      (setq file (expand-file-name file dir)))
+
+    (downcase file))
+
+   (t ;; linux
+    (when (and dir
+	       (not ada-xref-full-path))
+      (setq file (expand-file-name file dir)))
+    file))
+  )
+
+;;;;; wisi-xref methods
+
+(cl-defstruct (gpr-query-xref (:include gnat-compiler))
+  ;; no new slots
+  )
+
+;;;###autoload
+(cl-defun create-gpr_query-xref
+    (&key
+     gpr-file
+     run-buffer-name
+     project-path
+     target
+     runtime
+     gnat-stub-opts
+     gnat-stub-cargs)
+  ;; See note on `create-ada-prj' for why this is not a defalias.
+  (make-gpr-query-xref
+   :gpr-file gpr-file
+   :run-buffer-name run-buffer-name
+   :project-path project-path
+   :target target
+   :runtime runtime
+   :gnat-stub-opts gnat-stub-opts
+   :gnat-stub-cargs gnat-stub-cargs
+   ))
+
+(cl-defmethod wisi-xref-parse-one ((xref gpr-query-xref) project name value)
+  (wisi-compiler-parse-one xref project name value))
+
+(cl-defmethod wisi-xref-parse-final ((xref gpr-query-xref) _project prj-file-name)
+  (setf (gnat-compiler-run-buffer-name xref) (gnat-run-buffer-name prj-file-name gpr-query-buffer-name-prefix))
+  (unless (gnat-compiler-gpr-file xref)
+    (if (string= "gpr" (file-name-extension prj-file-name))
+	(setf (gnat-compiler-gpr-file xref) prj-file-name)
+      (user-error "using gpr-query xref, but no gpr file provided"))))
+
+(cl-defmethod wisi-xref-refresh-cache ((_xref gpr-query-xref) project no-full)
   ;; Kill the current session and delete the database, to get changed
   ;; env vars etc when it restarts.
   ;;
   ;; We need to delete the database files if the compiler version
   ;; changed, or the database was built with an incorrect environment
   ;; variable, or something else screwed up. However, rebuilding after
-  ;; that is a lot slower, so we only do that on request.
-  (let* ((session (gpr-query-cached-session))
+  ;; that is a lot slower, so we only do that with permission.
+  (let* ((session (gpr-query-cached-session project))
 	 (db-filename
-	  (with-current-buffer (gpr-query-session-send "db_name" t)
+	  (with-current-buffer (gpr-query-session-send session "db_name" t)
 	    (goto-char (point-min))
 	    (buffer-substring-no-properties (line-beginning-position) (line-end-position)))))
 
     (gpr-query-kill-session session)
-    (when delete-files
+    (unless no-full
       (delete-file db-filename))
-    (gpr-query--start-process session)))
 
-(defun gpr-query-other (identifier file line col)
-  "For `ada-xref-other-function', using gpr_query.
-FILE must be non-nil; line, col can be nil."
+    ;; We do this here because the user expects to see the "running
+    ;; gpr_query ..." message now, not later when a gpr-query command
+    ;; is run.
+    (gpr-query--start-process project session)
+    ))
+
+(cl-defmethod wisi-xref-other ((_xref gpr-query-xref) project &key identifier filename line column)
   (when (eq ?\" (aref identifier 0))
     ;; gpr_query wants the quotes stripped
-    (setq col (+ 1 col))
+    (setq column (+ 1 column))
     (setq identifier (substring identifier 1 (1- (length identifier))))
     )
 
-  (setq file (gpr-query--normalize-filename file))
+  (setq filename (gpr-query--normalize-filename filename))
 
   (let ((cmd (format "refs %s:%s:%s:%s"
 		     identifier
-		     (file-name-nondirectory file)
+		     (file-name-nondirectory filename)
 		     (or line "")
-		     (if col (1+ col) "")))
+		     (if column (1+ column) "")))
 	(decl-loc nil)
 	(body-loc nil)
 	(search-type nil)
 	(min-distance most-positive-fixnum)
-	(result nil))
+	(result nil)
+	(session (gpr-query-cached-session project)))
 
-    (with-current-buffer (gpr-query-session-send cmd t)
+    (with-current-buffer (gpr-query-session-send session cmd t)
       ;; 'gpr_query refs' returns a list containing the declaration,
       ;; the body, and all the references, in no particular order.
       ;;
@@ -503,6 +543,8 @@ FILE must be non-nil; line, col can be nil."
       ;; Module_Type:/home/Projects/GDS/work_stephe_2/common/1553/gds-hardware-bus_1553-wrapper.ads:171:9 (full declaration)
       ;;
       ;; itc_assert:/home/Projects/GDS/work_stephe_2/common/itc/opsim/itc_dscovr_gdsi/Gds1553/src/Gds1553.cpp:830:9 (reference)
+      ;;
+      ;; if ada-xref-full-path is nil, only the file basename is present.
 
       (message "parsing result ...")
 
@@ -516,12 +558,12 @@ FILE must be non-nil; line, col can be nil."
 		 (found-line (string-to-number (match-string 2)))
 		 (found-col  (string-to-number (match-string 3)))
 		 (found-type (match-string 4))
-		 (dist       (if (and line col)
-				 (gpr-query-dist found-line line found-col col)
+		 (dist       (if (and line column)
+				 (gpr-query-dist found-line line found-col column)
 			       most-positive-fixnum))
 		 )
 
-            (setq found-file (gpr-query--normalize-filename found-file))
+            (setq found-file (gpr-query--normalize-filename found-file (file-name-directory filename)))
 
 	    (cond
 	     ((string-equal found-type "declaration")
@@ -533,7 +575,7 @@ FILE must be non-nil; line, col can be nil."
 	      (setq body-loc (list found-file found-line (1- found-col))))
 	     )
 
-	    (when (and (equal found-file file)
+	    (when (and (equal found-file filename)
 		       (or
 			(string-equal found-type "body")
 			(string-equal found-type "declaration"))
@@ -575,7 +617,7 @@ FILE must be non-nil; line, col can be nil."
 	;; We started in the spec file; find the declaration
 	;;
 	;; If the file has both declaration and body, this will go to
-	;; declaration. Then a search with line, col can go to body.
+	;; declaration. Then a search with line, column can go to body.
 	(setq result decl-loc))
 
        ((and
@@ -594,30 +636,26 @@ FILE must be non-nil; line, col can be nil."
       (message "parsing result ... done")
       result)))
 
-(defun gpr-query-all (identifier file line col &optional _local-only _append)
-  "For `ada-xref-all-function', using gpr_query."
-  ;; FIXME: implement local-only, append
-  (gpr-query-compilation identifier file line col "refs" 'gpr-query-ident-file))
+(cl-defmethod wisi-xref-parents ((_xref gpr-query-xref) project &key identifier filename line column)
+  (gpr-query-compilation project identifier filename line column "parent_types" 'gpr-query-ident-file))
 
-(defun gpr-query-parents (identifier file line col)
-  "For `ada-xref-parent-function', using gpr_query."
-  (gpr-query-compilation identifier file line col "parent_types" 'gpr-query-ident-file))
+(cl-defmethod wisi-xref-all ((_xref gpr-query-xref) project &key identifier filename line column local-only append)
+  (gpr-query-compilation project identifier filename line column "refs" 'gpr-query-ident-file local-only append))
 
-(defun gpr-query-overriding (identifier file line col)
-  "For `ada-xref-overriding-function', using gpr_query."
-  (gpr-query-compilation identifier file line col "overriding" 'gpr-query-ident-file))
+(cl-defmethod wisi-xref-overriding ((_xref gpr-query-xref) project &key identifier filename line column)
+  (gpr-query-compilation project identifier filename line column "overriding" 'gpr-query-ident-file))
 
-(defun gpr-query-overridden-1 (identifier file line col)
-  "For `ada-xref-overridden-function', using gpr_query."
+(cl-defmethod wisi-xref-overridden ((_xref gpr-query-xref) project &key identifier filename line column)
   (when (eq ?\" (aref identifier 0))
     ;; gpr_query wants the quotes stripped
-    (setq col (+ 1 col))
+    (setq column (+ 1 column))
     (setq identifier (substring identifier 1 (1- (length identifier))))
     )
 
-  (let ((cmd (format "overridden %s:%s:%d:%d" identifier (file-name-nondirectory file) line (1+ col)))
+  (let ((cmd (format "overridden %s:%s:%d:%d" identifier (file-name-nondirectory filename) line (1+ column)))
+	(session (gpr-query-cached-session project))
 	result)
-    (with-current-buffer (gpr-query-session-send cmd t)
+    (with-current-buffer (gpr-query-session-send session cmd t)
 
       (goto-char (point-min))
       (when (looking-at gpr-query-ident-file-regexp)
@@ -633,75 +671,8 @@ FILE must be non-nil; line, col can be nil."
       (message "parsing result ... done")
       result)))
 
-(defun gpr-query--normalize-filename (file)
-  "Takes account of filesystem differences."
-  (when (eq system-type 'windows-nt)
-    ;; 'expand-file-name' converts Windows directory
-    ;; separators to normal Emacs.  Since Windows file
-    ;; system is case insensitive, GNAT and Emacs can
-    ;; disagree on the case, so convert all to lowercase.
-    (setq file (downcase (expand-file-name file))))
-  (when (eq system-type 'darwin)
-    ;; case-insensitive case-preserving; so just downcase
-    (setq file (downcase file)))
-  file
-  )
-
-(defun ada-gpr-query-select-prj ()
-  ;; We wait until the session is actually required to create it.
-  (setq ada-file-name-from-ada-name 'ada-gnat-file-name-from-ada-name)
-  (setq ada-ada-name-from-file-name 'ada-gnat-ada-name-from-file-name)
-  (setq ada-make-package-body       'ada-gnat-make-package-body)
-
-  (setq ada-xref-refresh-function    'gpr-query-refresh)
-  (setq ada-xref-all-function        'gpr-query-all)
-  (setq ada-xref-other-function      'gpr-query-other)
-  (setq ada-xref-parent-function     'gpr-query-parents)
-  (setq ada-xref-all-function        'gpr-query-all)
-  (setq ada-xref-overriding-function 'gpr-query-overriding)
-  (setq ada-xref-overridden-function 'gpr-query-overridden-1)
-  (setq ada-show-xref-tool-buffer    'gpr-query-show-buffer)
-
-  (when (fboundp 'xref-ada-mode)
-    (xref-ada-mode 1))
-
-  (add-to-list 'completion-ignored-extensions ".ali") ;; gnat library files, used for cross reference
-  )
-
-(defun ada-gpr-query-deselect-prj ()
-  ;; We don’t kill the session here; user may switch back to this
-  ;; project.
-  (setq ada-file-name-from-ada-name nil)
-  (setq ada-ada-name-from-file-name nil)
-  (setq ada-make-package-body       nil)
-
-  (setq ada-xref-other-function      nil)
-  (setq ada-xref-parent-function     nil)
-  (setq ada-xref-all-function        nil)
-  (setq ada-xref-overriding-function nil)
-  (setq ada-xref-overridden-function nil)
-  (setq ada-show-xref-tool-buffer    nil)
-
-  (when (fboundp 'xref-ada-mode)
-    (xref-ada-mode 0))
-
-  (setq completion-ignored-extensions (delete ".ali" completion-ignored-extensions))
-  )
-
-(defun ada-gpr-query ()
-  "Set Ada mode global vars to use gpr_query."
-  (add-to-list 'ada-prj-parser-alist       '("gpr" . gnat-parse-gpr))
-  (add-to-list 'ada-select-prj-xref-tool   '(gpr_query . ada-gpr-query-select-prj))
-  (add-to-list 'ada-deselect-prj-xref-tool '(gpr_query . ada-gpr-query-deselect-prj))
-
-  ;; no parse-*-xref
-  )
-
-(provide 'gpr-query)
-
 (add-to-list 'compilation-error-regexp-alist-alist
 	     (cons 'gpr-query-ident-file gpr-query-ident-file-regexp-alist))
 
-(ada-gpr-query)
-
+(provide 'gpr-query)
 ;;; end of file
